# These functions merge SAGE library files and write the merged data
# to a database table.
# libNames - a character string or a vector of character strings for
# the name of a directory where sage lib files are contained or for
# sage library files.
# infoName - a character string for the name of the file that keeps
# the mapping bettwen the original file names and a code in the form
# of "lib1", "lib2", ... . The mappings can be used to map table
# columns back to original file names when a database table is created
# using the merged file.
# isDir - a boolean to indicate whether libNames is only a string for
# the name of the directory where the original files are stored.
# skip - an integer for the number of top lines to be skiped from each
# SAGE library when reading the file.
# pattern - a character string for the pattern shared by all the original
# files to be merged. This argument is useful when only files with the given
# pattern will be merged.
#
# Copyright 2003, J. Zhang, all rights reserved.
#

mergeSAGE <- function(libNames, isDir = TRUE,  skip = 1, pattern = ".sage"){
    if(isDir){
        if(!missing(pattern)){
            fileNames <- list.files(path = libNames, pattern = pattern,
                          full.name = TRUE)
        }else{
            fileNames <- list.files(path = libNames, full.name = TRUE)
        }
    }else{
        fileNames <- libNames
    }
    path <- file.path(.path.package("SAGElyzer"), "temp")
    # Get the unique tags
    # print("Finding the unique tags")
    uniqTags <- getUniqTags(fileNames, skip = skip)
    uTagFile <- file.path(path, basename(tempfile("uTag")))
    write.table(uniqTags, uTagFile, quote = FALSE,
                sep = "\t", row.names = FALSE, col.names = FALSE)
    # Sort so that correct mappings between lib code and original lib
    # files can be obtained
    fileNames <- sort(fileNames)
    names(fileNames) <- paste("lib", 1:length(fileNames), sep = "")
    # Creating individual files each has count for each of the unique
    # tags using the original files. Tags missing from an original
    # file have a count of 0
    # print("Processing and merging files")
    writeSAGECounts(fileNames, uniqTags, skip)
    # A file name for the output file
    outName <-  file.path(path, "sage.merged")
    # Write the information data file
    infoName <- file.path(path, "sage.info")
    write.table(getLibInfo(fileNames), infoName, quote = FALSE,
                sep = "\t", row.names = FALSE, col.names = FALSE)

    if(.Platform$OS.type == "unix"){
        # Write to a file using unix command - paste
        system(paste(c("paste", uTagFile,
                       file.path(path, paste(names(fileNames),
                                       ".out", sep = "")), ">",
                                               outName), collapse = " "))
        writeSAGE4Unix(outName, infoName)
    }else if(.Platform$OS.type == "windows"){
        writeSAGE4Win(fileNames, uniqTags, infoName)
    }
    # Remove the temporary files
    file.remove(c(uTagFile, file.path(path, paste(names(fileNames),
                                       ".out", sep = ""))))
}

# This function returns a data frame with three columns for the database
# table column name, library file name, and count of number of rows for
# each SAGE library.
getLibInfo <- function(fileNames){

#    names(fileNames) <- paste("lib", 1:length(fileNames), sep = "")

    libinfo <- matrix(ncol = 4, nrow = length(fileNames))
    libinfo[,1] <- names(fileNames)
    libinfo[,2] <- basename(fileNames)
    libinfo[,3] <- calNormFact("min", getLibNNum(fileNames))
    libinfo[,4] <- calNormFact("max", getLibNNum(fileNames))

    return(libinfo)
}

# This function calculates the normalization factors for SAGE
# libraries. libNNum is a named vactor with the value being the number
# of row for each name.
calNormFact <- function(normalize = c("min", "max"), libNNum){

    normalize <- match.arg(normalize)
    switch(normalize,
           min = temp <- min(as.numeric(libNNum)),
           max = temp <- max(as.numeric(libNNum)),
           stop("Invalid argument for normalize"))

    return(signif(as.numeric(temp/as.numeric(libNNum)) ,4))
}

# This function calculates the number of rows for each of the files
# passed as an argument. Argument fileNames must be a named vector.
getLibNNum <- function(fileNames){
    getNum <- function(fileName){
        return(length(readLines(fileName)))
    }
    return(sapply(fileNames, getNum))
}

# This function reads SAGE libraries and extracts the unique tags from
# them.
# fileNames - a character string for sage library files.
# skip - an integer to indicate how many lines of each of the library
# files to be skiped.

getUniqTags <- function(fileNames, skip = 1, sep = "\t"){

    tagFromOne <- function(filename){
        temp <- read.table(filename, header = FALSE, skip = skip,
                           sep = sep, as.is = TRUE)
        return(temp[,1])
    }

    uniqueTags <- sapply(fileNames, tagFromOne)
    # Do it by steps to avoid memory problems in Windows
    uniqueTags <- unique(unlist(uniqueTags, use.names = FALSE))
    # Remove entries that are not of correct length
    # nchr <- sapply(uniqueTags, nchar)
    # nchr <- max(nchr[duplicated(nchr)])
    # uniqueTags <- uniqueTags[sapply(uniqueTags, nchar) == nchr]

    return(uniqueTags)
}

# This function reads in a vector of uniquie SAGE tags and maps counts
# from a SAGE library to the correct cell.
writeSAGE4Win <- function(fileNames, uniqTags, infoData, pace = 1000){

    path <- file.path(.path.package("SAGElyzer"), "temp")

    dbArgs <- getSLyzerArgs()$dbArgs
    conn <- makeConnection(dbArgs)
    if(tableExists(conn, dbArgs[["counts"]])){
        executeQuery(paste("drop table", dbArgs[["counts"]]), conn, TRUE)
    }
    infoFile <- read.table(infoData, header = FALSE, sep = "\t", as.is = TRUE)
    dbArgs <- getSLyzerArgs()$dbArgs
    colNames <- c("tag", infoFile[,1])
    keys <- "tag"
    intCols <- colNames[colNames != "tag"]
    # Create empty tables for counts
    newTableSQL <- paste("create table ", dbArgs[["counts"]], " (",
                         getColSQL(colNames, 21, "tag", infoFile[,1], "int4"),
                         "primary key(tag))", sep = "")
    executeQuery(newTableSQL, conn, TRUE)
    nrow <- length(uniqTags)
    begin <- 1
    end <- pace
    skip <- 0
    # Chop off a section from uniqTags, map file to tags, and write to file
    for(j in 1:floor(nrow/pace)){
        # Force a clean up
        temp <- mapFile2Tag(file.path(path, paste(names(fileNames),
                     ".out", sep = "")), uniqTags[begin:end], skip, pace)
        # Write to a table
        temp <- as.data.frame(temp)
        colnames(temp) <- colNames
        sqlSave(conn, temp, dbArgs[["counts"]], TRUE, FALSE, FALSE)

        begin <- end + 1
        skip <- end
        end <- (j + 1) * pace
    }
    temp <- mapFile2Tag(file.path(path, paste(names(fileNames),
               ".out", sep = "")), uniqTags[begin:length(uniqTags)],
                        skip, length(uniqTags) - skip)
    temp <- as.data.frame(temp)
    colnames(temp) <- colNames
    sqlSave(conn, temp, dbArgs[["counts"]], TRUE, FALSE, FALSE)
    temp <- NULL
    # Write info to table
    if(tableExists(conn, dbArgs[["info"]])){
        executeQuery(paste("drop table", dbArgs[["info"]]), conn, TRUE)
    }
    colNames <- c("libname", "filename", "minfactor", "maxfactor")
    intCols <- c("minfactor", "maxfactor")
    newTableSQL <- paste("create table ", dbArgs[["counts"]], " (",
                         getColSQL(colNames, 40, "libname",
                                   intCols, "float4"),
                         "primary key(libname))", sep = "")
    executeQuery(newTableSQL, conn, TRUE)
    temp <- read.table(infoData, sep = "\t", header = FALSE, as.is = TRUE)
    colnames(temp) <- colNames
    sqlSave(conn, temp, dbArgs[["info"]], TRUE, FALSE, FALSE)

#    closeConn(conn)
}

writeSAGE4Unix <- function(countData, infoData){

    infoFile <- read.table(infoData, header = FALSE, sep = "\t", as.is = TRUE)
    dbArgs <- getSLyzerArgs()$dbArgs
    # Write counts to table
    colNames <- c("tag", infoFile[,1])
    keys <- "tag"
    intCols <- colNames[colNames != "tag"]
    writeSAGE2DB(dbArgs, colNames, keys, intCols, countData, "counts")
    # Write info to table
    colNames <- c("libname", "filename", "minfactor", "maxfactor")
        keys <- "libname"
    intCols <- c("minfactor", "maxfactor")
    writeSAGE2DB(dbArgs, colNames, keys, intCols, infoData,
                     "info", charNum = 40, type = "float4")
}

# Map counts from each file to a chunk of unique tags
mapFile2Tag <- function(fileNames, tags, skip, n){
    mapOne <- function(filename){
        temp <- scan(filename, what = "", skip = skip, n = n,quiet = TRUE)
        return(temp)
    }

    mapped <- sapply(fileNames, mapOne)
    mapped <- matrix(unlist(mapped, use.names = FALSE), byrow = FALSE,
                     ncol = length(fileNames))
    mapped <- cbind(matrix(tags, ncol = 1), mapped)

    return(mapped)
}

# This function reads the library files, map counts to unique SAGE
# tags, and create files containg the counts only. Tags missing from
# the original files will have counts 0.
writeSAGECounts <- function(fileNames, uniqTags, skip, sep = "\t"){
#    outNames <- NULL
    # Process each library and write counts from a library to the
    # correct location
    for(i in 1:length(fileNames)){
        # Read in a library
        temp <- as.matrix(read.table(fileNames[i], header = FALSE,
                                 sep = sep, skip = skip, as.is = TRUE))

        # Match unique tags agaist the library
        out <- rep(0, length(uniqTags))
#        matched <- match(temp[,1], uniqTags)
        out[match(temp[,1], uniqTags)] <- gsub(" *", "", temp[,2])
        # Free memory
        temp = NULL
        gc()
        # Write to a file
        tempName <- file.path(.path.package("SAGElyzer"), "temp",
                              paste(names(fileNames)[i], ".out",
                                    sep = ""))
        write(out, ncol=1, file = tempName)
#        outNames <- c(outNames, tempName)
    }
#    return(outNames)
}










